# Copyright      2024 Wei Kang (wkang@pku.edu.cn)
#
# See ../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import _ssentencepiece
import os
import unicodedata
from typing import List, Union

import importlib_resources


class Ssentencepiece:
    def __init__(
        self, vocab: str = "librispeech-500", num_threads: int = os.cpu_count()
    ):
        """
        Construct a Ssentencepiece object.

        Args:
          vocab:
            The path of vocab file (this vocab file must be created by google's
            sentencepiece trainer.
          num_threads:
            The number of worker threads when encode/decode multiple sequences.
            Default `os.cpu_count()`.
        """
        if not os.path.exists(vocab):
            ref = (
                importlib_resources.files("ssentencepiece")
                / f"resources/{vocab}.vocab"
            )
            with importlib_resources.as_file(ref) as path:
                vocab = str(path)
            assert os.path.exists(vocab), f"Cannot find {vocab} file."

        self.vocab = vocab
        self.num_threads = num_threads
        self.processor = _ssentencepiece.ssentencepiece(vocab, num_threads)

    def __getstate__(self):
        # serialize only picklable parts
        return {"vocab": self.vocab, "num_threads": self.num_threads}

    def __setstate__(self, state):
        # restore picklable parts
        self.vocab = state["vocab"]
        self.num_threads = state["num_threads"]
        # lazily recreate non-picklable parts
        self.processor = _ssentencepiece.ssentencepiece(
            self.vocab, self.num_threads
        )

    def encode(
        self, text: Union[str, List[str]], out_type=int
    ) -> Union[List[Union[int, str]], List[List[Union[int, str]]]]:
        """
        Encode the text into tokens or token ids, almost the same as the encode
        interface in sentencepiece.

        Args:
          text:
            The input text, could be a single string or a list of strings.
          out_type:
            Whether to output tokens (out_type=str) or token ids(out_type=int).

        Return:
          If text is a single string and out_type is str, outputs a list of strings.
          If text is a single string and out_type is int, outputs a list of ints.
          If text is a list of strings and out_type is str, outputs a list of list of strings.
          If text is a list of strings and out_type is int, outputs a list of list of ints.
        """
        if out_type is int:
            return self.processor.encode(text, output_id=True)
        else:
            assert (
                out_type is str
            ), f"The out_type could only be int or str, given {out_type}"
            # google's sentencepiece uses NFKC normalization.
            if isinstance(text, str):
                text = unicodedata.normalize("NFKC", text)
            else:
                text = [unicodedata.normalize("NFKC", x) for x in text]
            return self.processor.encode(text, output_id=False)

    def decode(
        self, ids: Union[List[int], List[List[int]]]
    ) -> Union[str, List[str]]:
        """
        Decode the token ids into string, almost the same as the decode interface in sentencepiece.

        Args:
          ids:
            The input ids, could be a list of ints of a list of a list of ints.

        Return:
          If the ids is a list of ints, outputs a single string.
          If the ids is a list of a list of ints, outputs a list of strings.
        """
        return self.processor.decode(ids)

    def vocab_size(self) -> int:
        """
        Return the vocabulary size.`
        """
        return self.processor.vocab_size()

    def id_to_piece(self, ids: Union[int, List[int]]) -> Union[str, List[str]]:
        """
        Convert the token id into piece.

        Args:
          ids:
            The input ids, could be a int of list of ints.

        Return:
          If the ids is a a int, outputs a single string.
          If the ids is a list of ints, outputs a list of strings.
        """
        return self.processor.id_to_piece(ids)

    def piece_to_id(
        self, pieces: Union[str, List[str]]
    ) -> Union[int, List[int]]:
        """
        Convert the piece into token id.

        Args:
          pieces:
            The input pieces, could be a list of strings of a list of a list of ints.

        Return:
          If the ids is a string, outputs a single int.
          If the ids is a list of string, outputs a list of ints.
        """

        # google's sentencepiece uses NFKC normalization.
        if isinstance(pieces, str):
            pieces = unicodedata.normalize("NFKC", pieces)
        else:
            pieces = [unicodedata.normalize("NFKC", x) for x in pieces]
        return self.processor.piece_to_id(pieces)



#!/usr/bin/env python3
#
# Copyright      2024 Wei Kang (wkang@pku.edu.cn)
#
# See ../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# To run this single test, use
#
#  ctest --verbose -R ssentencepiece_test_py


import unittest
import numpy as np

from ssentencepiece import Ssentencepiece
import sentencepiece as spm


class TestEncodeDecode(unittest.TestCase):
    def test_encode_decode(self):
        sp = spm.SentencePieceProcessor()
        sp.load("ssentencepiece/python/tests/testdata/bpe.model")
        ssp = Ssentencepiece("ssentencepiece/python/tests/testdata/bpe.vocab")
        sentences = [
            "Licensed under the Apache License Version the License",
            "you may not use this file except in compliance with the License",
            "You may obtain a copy of the License at",
            "Unless required by applicable law or agreed to in writing software",
            "distributed under the License is distributed on an AS IS BASIS",
            "WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND either express or implied",
            "See the License for the specific language governing permissions and",
        ]
        sentences = [x.upper() for x in sentences]
        for sen in sentences:
            r1 = sp.encode(sen)
            r2 = ssp.encode(sen)
            assert r1 == r2, (r1, r2)
            r1 = sp.encode(sen, out_type=str)
            r2 = ssp.encode(sen, out_type=str)
            assert r1 == r2, (r1, r2)
            r1 = sp.decode(sp.encode(sen))
            r2 = ssp.decode(ssp.encode(sen))
            assert r1 == r2, (r1, r2)

        r1 = sp.encode(sentences)
        r2 = ssp.encode(sentences)
        assert r1 == r2, (r1, r2)
        r1 = sp.decode(sp.encode(sentences))
        r2 = ssp.decode(ssp.encode(sentences))
        assert r1 == r2, (r1, r2)

    def test_encode_decode_bbpe(self):
        sp = spm.SentencePieceProcessor()
        sp.load("ssentencepiece/python/tests/testdata/bbpe.model")
        ssp = Ssentencepiece("ssentencepiece/python/tests/testdata/bbpe.vocab")
        sentences = [
            "Licensed under the Apache License Version the License",
            "you may not use this file except in compliance with the License",
            "You may obtain a copy of the License at",
            "Unless required by applicable law or agreed to in writing software",
            "distributed under the License is distributed on an AS IS BASIS",
            "WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND either express or implied",
            "See the License for the specific language governing permissions and",
        ]
        sentences = [x.upper() for x in sentences]
        for sen in sentences:
            r1 = sp.encode(sen)
            r2 = ssp.encode(sen)
            assert r1 == r2, (r1, r2)
            r1 = sp.encode(sen, out_type=str)
            r2 = ssp.encode(sen, out_type=str)
            assert r1 == r2, (r1, r2)
            r1 = sp.decode(sp.encode(sen))
            r2 = ssp.decode(ssp.encode(sen))
            assert r1 == r2, (r1, r2)

        r1 = sp.encode(sentences)
        r2 = ssp.encode(sentences)
        assert r1 == r2, (r1, r2)
        r1 = sp.decode(sp.encode(sentences))
        r2 = ssp.decode(ssp.encode(sentences))
        assert r1 == r2, (r1, r2)

    def test_id_to_piece_to_id(self):
        ssp = Ssentencepiece("ssentencepiece/python/tests/testdata/bpe.vocab")
        for i in range(ssp.vocab_size()):
            piece = ssp.id_to_piece(i)
            id = ssp.piece_to_id(piece)
            assert i == id, (i, piece, id)


if __name__ == "__main__":
    unittest.main()



# Copyright      2025 Wei Kang (wkang@pku.edu.cn)
#
# See ../../../LICENSE for clarification regarding multiple authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import click
import logging
import sys

from pathlib import Path
from ssentencepiece import Ssentencepiece

try:
    import sentencepiece as spm
except ImportError:
    print("Please run")
    print("  pip install sentencepiece")
    print("before you continue")
    raise


@click.group()
def cli():
    """
    The shell entry point to ssentencepiece.
    """
    logging.basicConfig(
        format="%(asctime)s %(levelname)s [%(filename)s:%(lineno)d] %(message)s",
        level=logging.INFO,
    )


@cli.command(name="export")
@click.option(
    "--bpe-model", type=str, required=True, help="The path to the bpe model."
)
def export(bpe_model):
    """
    Export the vocabulary (needed by simple-sentencepiece) from a BPE model (trained with google's sentencepiece).
    The vocabulary is written to a file with the same name as the model but with a .vocab extension.
    """
    model_file = bpe_model
    vocab_file = model_file.replace(".model", ".vocab")

    sp = spm.SentencePieceProcessor()
    sp.Load(model_file)
    vocabs = [sp.IdToPiece(id) for id in range(sp.GetPieceSize())]
    with open(vocab_file, "w") as vfile:
        for v in vocabs:
            id = sp.PieceToId(v)
            vfile.write(f"{v}\t{sp.GetScore(id)}\n")
    logging.info(f"Vocabulary file is written to {vocab_file}")


@cli.command(name="train")
@click.option(
    "--output-dir",
    type=str,
    required=True,
    help="The directory to store the trained model",
)
@click.option(
    "--texts",
    type=str,
    required=True,
    help="The path to the texts that will be used to train the model.",
)
@click.option(
    "--vocab-size",
    type=int,
    default=5000,
    help="The size of the vocabulary.",
)
@click.option(
    "--character-coverage",
    type=float,
    default=1.0,
    help="The character coverage for the model.",
)
@click.option(
    "--model-type",
    type=click.Choice(["unigram", "char", "word"]),
    default="unigram",
    help="The type of the model.",
)
@click.option(
    "--input-sentence-size",
    type=int,
    default=100000000,
    help="The number of sentences to use for training.",
)
def train_bpe_model(
    output_dir,
    texts,
    vocab_size,
    character_coverage,
    model_type,
    input_sentence_size,
):
    """
    Use the sentencepiece library to train a BPE model.
    The first 5 tokens are fixed to <blk>, <sos>, <eos>, <pad> and <unk>.
    """
    model_prefix = f"{output_dir}/{model_type}_{vocab_size}"

    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    user_defined_symbols = ["<blk>", "<sos>", "<eos>", "<pad>"]
    unk_id = len(user_defined_symbols)
    # Note: unk_id is fixed to 4.
    # If you change it, you should also change other
    # places that are using it.

    model_file = Path(model_prefix + ".vocab")
    if not model_file.is_file():
        spm.SentencePieceTrainer.train(
            input=texts,
            vocab_size=vocab_size,
            model_type=model_type,
            model_prefix=model_prefix,
            input_sentence_size=input_sentence_size,
            character_coverage=character_coverage,
            user_defined_symbols=user_defined_symbols,
            unk_id=unk_id,
            bos_id=-1,
            eos_id=-1,
        )
    else:
        print(f"{model_file} exists - skipping")
        return


@cli.command(name="encode")
@click.option(
    "--bpe-model",
    type=str,
    help="""
    The path to the bpe model, if provided will use standard sentencepiece
    toolkit (from google) to encode the text.
    """,
)
@click.option(
    "--bpe-vocab",
    type=str,
    help="""The path to the bpe model, if provided will use simple-sentencepiece
    (current one) to encode the text. If provided it will override the bpe-model.
    """,
)
@click.option(
    "--texts",
    type=str,
    help="The path to the texts to encode. If not provided, will read from stdin.",
)
@click.option(
    "--output-format",
    type=click.Choice(["piece", "id"]),
    default="piece",
    help="The output format, either piece or id.",
)
@click.option(
    "--output",
    type=str,
    help="The path to the output file. If not provided, will write to stdout.",
)
def encode_texts(bpe_model, bpe_vocab, texts, output_format, output):
    """
    Encode the texts to tokens or token ids.
    """
    if bpe_vocab:
        sp = Ssentencepiece(bpe_vocab)
    else:
        assert bpe_model, "Either bpe-model or bpe-vocab should be provided"
        sp = spm.SentencePieceProcessor()
        sp.Load(bpe_model)
    if output:
        output = open(output, "w")
    else:
        output = sys.stdout

    out_type = int if output_format == "id" else str

    if texts:
        with open(texts, "r") as f:
            for sen in f:
                sen = sen.strip()
                r = sp.encode(sen, out_type=out_type)
                output.write(" ".join([str(rr) for rr in r]) + "\n")
    else:
        for sen in sys.stdin:
            sen = sen.strip()
            r = sp.encode(sen, out_type=out_type)
            output.write(" ".join([str(rr) for rr in r]) + "\n")


@cli.command(name="decode")
@click.option(
    "--bpe-model",
    type=str,
    help="""
    The path to the bpe model, if provided will use standard sentencepiece
    toolkit (from google) to encode the text.
    """,
)
@click.option(
    "--bpe-vocab",
    type=str,
    help="""The path to the bpe model, if provided will use simple-sentencepiece
    (current one) to encode the text. If provided it will override the bpe-model.
    """,
)
@click.option(
    "--input",
    type=str,
    help="The path to the pieces/ids to decode. If not provided, will read from stdin.",
)
@click.option(
    "--input-format",
    type=click.Choice(["piece", "id"]),
    default="piece",
    help="The input format, either piece or id.",
)
@click.option(
    "--output",
    type=str,
    help="The path to the output file. If not provided, will write to stdout.",
)
def decode_pieces(bpe_model, bpe_vocab, input, input_format, output):
    """
    Decode the tokens or token ids to texts.
    tokens and token ids are separated by space.
    """
    if bpe_vocab:
        sp = Ssentencepiece(bpe_vocab)
    else:
        assert bpe_model, "Either bpe-model or bpe-vocab should be provided"
        sp = spm.SentencePieceProcessor()
        sp.Load(bpe_model)
    if output:
        output = open(output, "w")
    else:
        output = sys.stdout

    if input:
        with open(input, "r") as f:
            for sen in f:
                sen = sen.strip().split()
                if input_format == "piece":
                    r = "".join(sen).replace("▁", " ").strip()
                else:
                    r = sp.decode([in_type(s) for s in sen])
                output.write(r + "\n")
    else:
        for sen in sys.stdin:
            sen = sen.strip().split()
            if input_format == "piece":
                r = "".join(sen).replace("▁", " ").strip()
            else:
                r = sp.decode([in_type(s) for s in sen])
            output.write(r + "\n")


@cli.command(name="combine")
@click.argument(
    "vocabs",
    nargs=-1,
    type=click.Path(exists=True, dir_okay=False, allow_dash=True),
)
@click.argument("output_vocab", type=click.Path(allow_dash=True))
def combine_models(vocabs, output_vocab):
    """
    Combine multiple bpe models into one.

    The same token in different vocabularies will be merged into one token (maximum score is selected).
    """
    special_tokens = {}
    byte_tokens = {}
    normal_tokens = {}
    for vocab in vocabs:
        with open(vocab, "r") as f:
            for index, line in enumerate(f):
                token, score = line.strip().split("\t")
                score = float(score)
                if token.startswith("<0x") and token.endswith(">"):
                    byte_tokens[token] = score
                elif token.startswith("<") and token.endswith(">"):
                    if token not in special_tokens:
                        special_tokens[token] = index
                    else:
                        special_tokens[token] = min(
                            special_tokens[token], index
                        )
                else:
                    if token not in normal_tokens:
                        normal_tokens[token] = score
                    else:
                        normal_tokens[token] = max(normal_tokens[token], score)

    with open(output_vocab, "w") as f:
        special_tokens = sorted(special_tokens.items(), key=lambda x: x[1])
        for token, _ in special_tokens:
            f.write(f"{token}\t{0}\n")
        byte_tokens = sorted(byte_tokens.items(), key=lambda x: x[0])
        for token, _ in byte_tokens:
            f.write(f"{token}\t{0}\n")
        normal_tokens = sorted(normal_tokens.items(), key=lambda x: x[0])
        for token, score in normal_tokens:
            f.write(f"{token}\t{score}\n")



#!/usr/bin/env python3
#
# Copyright (c)  2024  Xiaomi Corporation (author: Wei Kang)

import glob
import os
import platform
import re
import shutil
import sys

import setuptools
from setuptools.command.build_ext import build_ext

cur_dir = os.path.dirname(os.path.abspath(__file__))


def is_windows():
    return platform.system() == "Windows"


def cmake_extension(name, *args, **kwargs) -> setuptools.Extension:
    kwargs["language"] = "c++"
    sources = []
    return setuptools.Extension(name, sources, *args, **kwargs)


class BuildExtension(build_ext):
    def build_extension(self, ext: setuptools.extension.Extension):
        # build/temp.linux-x86_64-3.8
        build_dir = self.build_temp
        os.makedirs(build_dir, exist_ok=True)

        # build/lib.linux-x86_64-3.8
        os.makedirs(self.build_lib, exist_ok=True)

        # ss is short for simple sentencepiece
        ss_dir = os.path.dirname(os.path.abspath(__file__))

        cmake_args = os.environ.get("SS_CMAKE_ARGS", "")
        make_args = os.environ.get("SS_MAKE_ARGS", "")
        system_make_args = os.environ.get("MAKEFLAGS", "")

        if cmake_args == "":
            cmake_args = "-DCMAKE_BUILD_TYPE=Release -DSBPE_ENABLE_TESTS=OFF"
            cmake_args += f" -DCMAKE_INSTALL_PREFIX={self.build_lib} "

        if make_args == "" and system_make_args == "":
            make_args = " -j "

        if "PYTHON_EXECUTABLE" not in cmake_args:
            print(f"Setting PYTHON_EXECUTABLE to {sys.executable}")
            cmake_args += f" -DPYTHON_EXECUTABLE={sys.executable}"


        if is_windows():
            build_cmd = f"""
         cmake {cmake_args} -B {self.build_temp} -S {ss_dir}
         cmake --build {self.build_temp} --target install --config Release -- -m
            """
            print(f"build command is:\n{build_cmd}")
            ret = os.system(
                f"cmake {cmake_args} -B {self.build_temp} -S {ss_dir}"
            )
            if ret != 0:
                raise Exception("Failed to configure simple sentencepiece")

            ret = os.system(
                f"cmake --build {self.build_temp} --target install --config Release -- -m"  # noqa
            )
            if ret != 0:
                raise Exception("Failed to build and install simple sentencepiece")
        else:
            build_cmd = f"""
                cd {self.build_temp}

                cmake {cmake_args} {ss_dir}

                make {make_args} install/strip
            """
            print(f"build command is:\n{build_cmd}")

            ret = os.system(build_cmd)
            if ret != 0:
                raise Exception(
                    "\nBuild ssentencepiece failed. Please check the error "
                    "message.\n"
                    "You can ask for help by creating an issue on GitHub.\n"
                    "\nClick:\n"
                    "\thttps://github.com/pkufool/simple-sentencepiece/issues/new\n"  # noqa
                )

        lib_so = glob.glob(f"{build_dir}/lib/*.so*")
        for so in lib_so:
            print(f"Copying {so} to {self.build_lib}/")
            shutil.copy(f"{so}", f"{self.build_lib}/")

        # macos
        lib_so = glob.glob(f"{build_dir}/lib/*.dylib*")
        for so in lib_so:
            print(f"Copying {so} to {self.build_lib}/")
            shutil.copy(f"{so}", f"{self.build_lib}/")


def get_package_version():
    with open("CMakeLists.txt") as f:
        content = f.read()

    latest_version = re.search(r"set\(SBPE_VERSION (.*)\)", content).group(1)
    latest_version = latest_version.strip('"')
    return latest_version


setuptools.setup(
    package_dir={
        "ssentencepiece": "ssentencepiece/python/ssentencepiece",
    },
    packages=["ssentencepiece"],
    ext_modules=[cmake_extension("_ssentencepiece")],
    cmdclass={"build_ext": BuildExtension},
)

with open("ssentencepiece/python/ssentencepiece/__init__.py", "a") as f:
    f.write(f"__version__ = '{get_package_version()}'\n")

with open("ssentencepiece/python/ssentencepiece/__init__.py", "r") as f:
    lines = f.readlines()

with open("ssentencepiece/python/ssentencepiece/__init__.py", "w") as f:
    for line in lines:
        if "__version__" in line:
            f.write(line)
            break
        f.write(line)

